"""
This module provides utilities for performing Azure Blob Storage operations without requiring
the heavyweight azure-storage-blob library dependency
"""
from copy import deepcopy
import urllib
import logging

from mlflow_databricks_artifacts.utils import rest_utils

_logger = logging.getLogger(__name__)
_PUT_BLOCK_HEADERS = {
    "x-ms-blob-type": "BlockBlob",
}


def put_adls_file_creation(sas_url, headers):
    """
    Performs an ADLS Azure file create `Put` operation
    (https://docs.microsoft.com/en-us/rest/api/storageservices/datalakestoragegen2/path/create)
    :param sas_url: A shared access signature URL referring to the Azure ADLS server
                    to which the file creation command should be issued.
    :param headers: Additional headers to include in the Put request body
    """
    request_url = _append_query_parameters(sas_url, {"resource": "file"})

    request_headers = {}
    for name, value in headers.items():
        if _is_valid_adls_put_header(name):
            request_headers[name] = value
        else:
            _logger.debug("Removed unsupported '%s' header for ADLS Gen2 Put operation", name)

    with rest_utils.cloud_storage_http_request(
            "put", request_url, headers=request_headers
    ) as response:
        rest_utils.augmented_raise_for_status(response)


def patch_adls_file_upload(sas_url, data, position, headers, is_single):
    """
    Performs an ADLS Azure file create `Patch` operation
    (https://docs.microsoft.com/en-us/rest/api/storageservices/datalakestoragegen2/path/update)
    :param sas_url: A shared access signature URL referring to the Azure ADLS server
                    to which the file update command should be issued.
    :param data: Data to include in the Patch request body.
    :param position: Positional offset of the data in the Patch request
    :param headers: Additional headers to include in the Patch request body
    """
    new_params = {"action": "append", "position": str(position)}
    if is_single:
        new_params["flush"] = "true"
    request_url = _append_query_parameters(sas_url, new_params)

    request_headers = {}
    for name, value in headers.items():
        if _is_valid_adls_patch_header(name):
            request_headers[name] = value
        else:
            _logger.debug("Removed unsupported '%s' header for ADLS Gen2 Patch operation", name)

    with rest_utils.cloud_storage_http_request(
            "patch", request_url, data=data, headers=request_headers
    ) as response:
        rest_utils.augmented_raise_for_status(response)


def patch_adls_flush(sas_url, position, headers):
    """
    Performs an ADLS Azure file flush `Patch` operation
    (https://docs.microsoft.com/en-us/rest/api/storageservices/datalakestoragegen2/path/update)
    :param sas_url: A shared access signature URL referring to the Azure ADLS server
                    to which the file update command should be issued.
    :param position: The final size of the file to flush.
    :param headers: Additional headers to include in the Patch request body.
    """
    request_url = _append_query_parameters(sas_url, {"action": "flush", "position": str(position)})

    request_headers = {}
    for name, value in headers.items():
        if _is_valid_adls_put_header(name):
            request_headers[name] = value
        else:
            _logger.debug("Removed unsupported '%s' header for ADLS Gen2 Patch operation", name)

    with rest_utils.cloud_storage_http_request(
            "patch", request_url, headers=request_headers
    ) as response:
        rest_utils.augmented_raise_for_status(response)


def put_block(sas_url, block_id, data, headers):
    """
    Performs an Azure `Put Block` operation
    (https://docs.microsoft.com/en-us/rest/api/storageservices/put-block)

    :param sas_url: A shared access signature URL referring to the Azure Block Blob
                    to which the specified data should be staged.
    :param block_id: A base64-encoded string identifying the block.
    :param data: Data to include in the Put Block request body.
    :param headers: Additional headers to include in the Put Block request body
                    (the `x-ms-blob-type` header is always included automatically).
    """
    request_url = _append_query_parameters(
        sas_url, {"comp": "block", "blockid": block_id}
    )

    request_headers = deepcopy(_PUT_BLOCK_HEADERS)
    for name, value in headers.items():
        if _is_valid_put_block_header(name):
            request_headers[name] = value
        else:
            _logger.debug(
                "Removed unsupported '%s' header for Put Block operation", name
            )

    with rest_utils.cloud_storage_http_request(
        "put", request_url, data=data, headers=request_headers
    ) as response:
        rest_utils.augmented_raise_for_status(response)


def put_block_list(sas_url, block_list, headers):
    """
    Performs an Azure `Put Block List` operation
    (https://docs.microsoft.com/en-us/rest/api/storageservices/put-block-list)

    :param sas_url: A shared access signature URL referring to the Azure Block Blob
                    to which the specified data should be staged.
    :param block_list: A list of uncommitted base64-encoded string block IDs to commit. For
                       more information, see
                       https://docs.microsoft.com/en-us/rest/api/storageservices/put-block-list.
    :param headers: Headers to include in the Put Block request body.
    """
    request_url = _append_query_parameters(sas_url, {"comp": "blocklist"})
    data = _build_block_list_xml(block_list)

    request_headers = {}
    for name, value in headers.items():
        if _is_valid_put_block_list_header(name):
            request_headers[name] = value
        else:
            _logger.debug(
                "Removed unsupported '%s' header for Put Block List operation", name
            )

    with rest_utils.cloud_storage_http_request(
        "put", request_url, data=data, headers=request_headers
    ) as response:
        rest_utils.augmented_raise_for_status(response)


def _append_query_parameters(url, parameters):
    parsed_url = urllib.parse.urlparse(url)
    query_dict = dict(urllib.parse.parse_qsl(parsed_url.query))
    query_dict.update(parameters)
    new_query = urllib.parse.urlencode(query_dict)
    new_url_components = parsed_url._replace(query=new_query)
    new_url = urllib.parse.urlunparse(new_url_components)
    return new_url


def _build_block_list_xml(block_list):
    xml = '<?xml version="1.0" encoding="utf-8"?>\n<BlockList>\n'
    for block_id in block_list:
        # Because block IDs are base64-encoded and base64 strings do not contain
        # XML special characters, we can safely insert the block ID directly into
        # the XML document
        xml += "<Uncommitted>{}</Uncommitted>\n".format(block_id)
    xml += "</BlockList>"
    return xml


def _is_valid_put_block_list_header(header_name):
    """
    :return: True if the specified header name is a valid header for the Put Block List operation,
             False otherwise. For a list of valid headers, see https://docs.microsoft.com/en-us/
             rest/api/storageservices/put-block-list#request-headers and https://docs.microsoft.com/
             en-us/rest/api/storageservices/
             specifying-conditional-headers-for-blob-service-operations#Subheading1.
    """
    return header_name.startswith("x-ms-meta-") or header_name in {
        "Authorization",
        "Date",
        "x-ms-date",
        "x-ms-version",
        "Content-Length",
        "Content-MD5",
        "x-ms-content-crc64",
        "x-ms-blob-cache-control",
        "x-ms-blob-content-type",
        "x-ms-blob-content-encoding",
        "x-ms-blob-content-language",
        "x-ms-blob-content-md5",
        "x-ms-encryption-scope",
        "x-ms-tags",
        "x-ms-lease-id",
        "x-ms-client-request-id",
        "x-ms-blob-content-disposition",
        "x-ms-access-tier",
        "If-Modified-Since",
        "If-Unmodified-Since",
        "If-Match",
        "If-None-Match",
    }


def _is_valid_put_block_header(header_name):
    """
    :return: True if the specified header name is a valid header for the Put Block operation, False
             otherwise. For a list of valid headers, see
             https://docs.microsoft.com/en-us/rest/api/storageservices/put-block#request-headers and
             https://docs.microsoft.com/en-us/rest/api/storageservices/put-block#
             request-headers-customer-provided-encryption-keys.
    """
    return header_name in {
        "Authorization",
        "x-ms-date",
        "x-ms-version",
        "Content-Length",
        "Content-MD5",
        "x-ms-content-crc64",
        "x-ms-encryption-scope",
        "x-ms-lease-id",
        "x-ms-client-request-id",
        "x-ms-encryption-key",
        "x-ms-encryption-key-sha256",
        "x-ms-encryption-algorithm",
    }


def _is_valid_adls_put_header(header_name):
    """
    :return: True if the specified header name is a valid header for the ADLS Put operation, False
             otherwise. For a list of valid headers, see
             https://docs.microsoft.com/en-us/rest/api/storageservices/datalakestoragegen2/path/create
    """
    return header_name in {
        "Cache-Control",
        "Content-Encoding",
        "Content-Language",
        "Content-Disposition",
        "x-ms-cache-control",
        "x-ms-content-type",
        "x-ms-content-encoding",
        "x-ms-content-language",
        "x-ms-content-disposition",
        "x-ms-rename-source",
        "x-ms-lease-id",
        "x-ms-properties",
        "x-ms-permissions",
        "x-ms-umask",
        "x-ms-owner",
        "x-ms-group",
        "x-ms-acl",
        "x-ms-proposed-lease-id",
        "x-ms-expiry-option",
        "x-ms-expiry-time",
        "If-Match",
        "If-None-Match",
        "If-Modified-Since",
        "If-Unmodified-Since",
        "x-ms-source-if-match",
        "x-ms-source-if-none-match",
        "x-ms-source-if-modified-since",
        "x-ms-source-if-unmodified-since",
        "x-ms-encryption-key",
        "x-ms-encryption-key-sha256",
        "x-ms-encryption-algorithm",
        "x-ms-encryption-context",
        "x-ms-client-request-id",
        "x-ms-date",
        "x-ms-version",
    }


def _is_valid_adls_patch_header(header_name):
    """
    :return: True if the specified header name is a valid header for the ADLS Patch operation, False
             otherwise. For a list of valid headers, see
             https://docs.microsoft.com/en-us/rest/api/storageservices/datalakestoragegen2/path/update
    """
    return header_name in {
        "Content-Length",
        "Content-MD5",
        "x-ms-lease-id",
        "x-ms-cache-control",
        "x-ms-content-type",
        "x-ms-content-disposition",
        "x-ms-content-encoding",
        "x-ms-content-language",
        "x-ms-content-md5",
        "x-ms-properties",
        "x-ms-owner",
        "x-ms-group",
        "x-ms-permissions",
        "x-ms-acl",
        "If-Match",
        "If-None-Match",
        "If-Modified-Since",
        "If-Unmodified-Since",
        "x-ms-encryption-key",
        "x-ms-encryption-key-sha256",
        "x-ms-encryption-algorithm",
        "x-ms-encryption-context",
        "x-ms-client-request-id",
        "x-ms-date",
        "x-ms-version",
    }


def chunk_list(l, chunk_size):
    for i in range(0, len(l), chunk_size):
        yield l[i : i + chunk_size]


def merge_dicts(dict_a, dict_b, raise_on_duplicates=True):
    """
    This function takes two dictionaries and returns one singular merged dictionary.

    :param dict_a: The first dictionary.
    :param dict_b: The second dictonary.
    :param raise_on_duplicates: If True, the function raises ValueError if there are duplicate keys.
                                Otherwise, duplicate keys in `dict_b` will override the ones in
                                `dict_a`.
    :return: A merged dictionary.
    """
    duplicate_keys = dict_a.keys() & dict_b.keys()
    if raise_on_duplicates and len(duplicate_keys) > 0:
        raise ValueError(
            f"The two merging dictionaries contains duplicate keys: {duplicate_keys}."
        )
    return {**dict_a, **dict_b}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: patched_databricks_artifacts.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import service as _service
from google.protobuf import service_reflection
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mlflow.protos.scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2
from mlflow.protos import databricks_pb2 as databricks__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"patched_databricks_artifacts.proto\x12\x1bmlflow_databricks_artifacts\x1a\x15scalapb/scalapb.proto\x1a\x10\x64\x61tabricks.proto\"\x89\x02\n\x16\x41rtifactCredentialInfo\x12\x0e\n\x06run_id\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x12\n\nsigned_uri\x18\x03 \x01(\t\x12O\n\x07headers\x18\x04 \x03(\x0b\x32>.mlflow_databricks_artifacts.ArtifactCredentialInfo.HttpHeader\x12\x41\n\x04type\x18\x05 \x01(\x0e\x32\x33.mlflow_databricks_artifacts.ArtifactCredentialType\x1a)\n\nHttpHeader\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\xaa\x02\n\x15GetCredentialsForRead\x12\x14\n\x06run_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x0c\n\x04path\x18\x02 \x03(\t\x12\x12\n\npage_token\x18\x03 \x01(\t\x1ax\n\x08Response\x12M\n\x10\x63redential_infos\x18\x02 \x03(\x0b\x32\x33.mlflow_databricks_artifacts.ArtifactCredentialInfo\x12\x17\n\x0fnext_page_token\x18\x03 \x01(\tJ\x04\x08\x01\x10\x02:_\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\xe2?1\n/com.databricks.mlflow.api.MlflowTrackingMessage\"\xab\x02\n\x16GetCredentialsForWrite\x12\x14\n\x06run_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x0c\n\x04path\x18\x02 \x03(\t\x12\x12\n\npage_token\x18\x03 \x01(\t\x1ax\n\x08Response\x12M\n\x10\x63redential_infos\x18\x02 \x03(\x0b\x32\x33.mlflow_databricks_artifacts.ArtifactCredentialInfo\x12\x17\n\x0fnext_page_token\x18\x03 \x01(\tJ\x04\x08\x01\x10\x02:_\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\xe2?1\n/com.databricks.mlflow.api.MlflowTrackingMessage*s\n\x16\x41rtifactCredentialType\x12\x11\n\rAZURE_SAS_URI\x10\x01\x12\x15\n\x11\x41WS_PRESIGNED_URL\x10\x02\x12\x12\n\x0eGCP_SIGNED_URL\x10\x03\x12\x1b\n\x17\x41ZURE_ADLS_GEN2_SAS_URI\x10\x04\x32\xb8\x03\n DatabricksMlflowArtifactsService\x12\xc6\x01\n\x15getCredentialsForRead\x12\x32.mlflow_databricks_artifacts.GetCredentialsForRead\x1a;.mlflow_databricks_artifacts.GetCredentialsForRead.Response\"<\xf2\x86\x19\x38\n4\n\x04POST\x12&/mlflow/artifacts/credentials-for-read\x1a\x04\x08\x02\x10\x00\x10\x03\x12\xca\x01\n\x16getCredentialsForWrite\x12\x33.mlflow_databricks_artifacts.GetCredentialsForWrite\x1a<.mlflow_databricks_artifacts.GetCredentialsForWrite.Response\"=\xf2\x86\x19\x39\n5\n\x04POST\x12\'/mlflow/artifacts/credentials-for-write\x1a\x04\x08\x02\x10\x00\x10\x03\x42,\n\x1f\x63om.databricks.api.proto.mlflow\x90\x01\x01\xa0\x01\x01\xe2?\x02\x10\x01')

_ARTIFACTCREDENTIALTYPE = DESCRIPTOR.enum_types_by_name['ArtifactCredentialType']
ArtifactCredentialType = enum_type_wrapper.EnumTypeWrapper(_ARTIFACTCREDENTIALTYPE)
AZURE_SAS_URI = 1
AWS_PRESIGNED_URL = 2
GCP_SIGNED_URL = 3
AZURE_ADLS_GEN2_SAS_URI = 4


_ARTIFACTCREDENTIALINFO = DESCRIPTOR.message_types_by_name['ArtifactCredentialInfo']
_ARTIFACTCREDENTIALINFO_HTTPHEADER = _ARTIFACTCREDENTIALINFO.nested_types_by_name['HttpHeader']
_GETCREDENTIALSFORREAD = DESCRIPTOR.message_types_by_name['GetCredentialsForRead']
_GETCREDENTIALSFORREAD_RESPONSE = _GETCREDENTIALSFORREAD.nested_types_by_name['Response']
_GETCREDENTIALSFORWRITE = DESCRIPTOR.message_types_by_name['GetCredentialsForWrite']
_GETCREDENTIALSFORWRITE_RESPONSE = _GETCREDENTIALSFORWRITE.nested_types_by_name['Response']
ArtifactCredentialInfo = _reflection.GeneratedProtocolMessageType('ArtifactCredentialInfo', (_message.Message,), {

  'HttpHeader' : _reflection.GeneratedProtocolMessageType('HttpHeader', (_message.Message,), {
    'DESCRIPTOR' : _ARTIFACTCREDENTIALINFO_HTTPHEADER,
    '__module__' : 'patched_databricks_artifacts_pb2'
    # @@protoc_insertion_point(class_scope:mlflow_databricks_artifacts.ArtifactCredentialInfo.HttpHeader)
    })
  ,
  'DESCRIPTOR' : _ARTIFACTCREDENTIALINFO,
  '__module__' : 'patched_databricks_artifacts_pb2'
  # @@protoc_insertion_point(class_scope:mlflow_databricks_artifacts.ArtifactCredentialInfo)
  })
_sym_db.RegisterMessage(ArtifactCredentialInfo)
_sym_db.RegisterMessage(ArtifactCredentialInfo.HttpHeader)

GetCredentialsForRead = _reflection.GeneratedProtocolMessageType('GetCredentialsForRead', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETCREDENTIALSFORREAD_RESPONSE,
    '__module__' : 'patched_databricks_artifacts_pb2'
    # @@protoc_insertion_point(class_scope:mlflow_databricks_artifacts.GetCredentialsForRead.Response)
    })
  ,
  'DESCRIPTOR' : _GETCREDENTIALSFORREAD,
  '__module__' : 'patched_databricks_artifacts_pb2'
  # @@protoc_insertion_point(class_scope:mlflow_databricks_artifacts.GetCredentialsForRead)
  })
_sym_db.RegisterMessage(GetCredentialsForRead)
_sym_db.RegisterMessage(GetCredentialsForRead.Response)

GetCredentialsForWrite = _reflection.GeneratedProtocolMessageType('GetCredentialsForWrite', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETCREDENTIALSFORWRITE_RESPONSE,
    '__module__' : 'patched_databricks_artifacts_pb2'
    # @@protoc_insertion_point(class_scope:mlflow_databricks_artifacts.GetCredentialsForWrite.Response)
    })
  ,
  'DESCRIPTOR' : _GETCREDENTIALSFORWRITE,
  '__module__' : 'patched_databricks_artifacts_pb2'
  # @@protoc_insertion_point(class_scope:mlflow_databricks_artifacts.GetCredentialsForWrite)
  })
_sym_db.RegisterMessage(GetCredentialsForWrite)
_sym_db.RegisterMessage(GetCredentialsForWrite.Response)

_DATABRICKSMLFLOWARTIFACTSSERVICE = DESCRIPTOR.services_by_name['DatabricksMlflowArtifactsService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\037com.databricks.api.proto.mlflow\220\001\001\240\001\001\342?\002\020\001'
  _GETCREDENTIALSFORREAD.fields_by_name['run_id']._options = None
  _GETCREDENTIALSFORREAD.fields_by_name['run_id']._serialized_options = b'\370\206\031\001'
  _GETCREDENTIALSFORREAD._options = None
  _GETCREDENTIALSFORREAD._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]\342?1\n/com.databricks.mlflow.api.MlflowTrackingMessage'
  _GETCREDENTIALSFORWRITE.fields_by_name['run_id']._options = None
  _GETCREDENTIALSFORWRITE.fields_by_name['run_id']._serialized_options = b'\370\206\031\001'
  _GETCREDENTIALSFORWRITE._options = None
  _GETCREDENTIALSFORWRITE._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]\342?1\n/com.databricks.mlflow.api.MlflowTrackingMessage'
  _DATABRICKSMLFLOWARTIFACTSSERVICE.methods_by_name['getCredentialsForRead']._options = None
  _DATABRICKSMLFLOWARTIFACTSSERVICE.methods_by_name['getCredentialsForRead']._serialized_options = b'\362\206\0318\n4\n\004POST\022&/mlflow/artifacts/credentials-for-read\032\004\010\002\020\000\020\003'
  _DATABRICKSMLFLOWARTIFACTSSERVICE.methods_by_name['getCredentialsForWrite']._options = None
  _DATABRICKSMLFLOWARTIFACTSSERVICE.methods_by_name['getCredentialsForWrite']._serialized_options = b'\362\206\0319\n5\n\004POST\022\'/mlflow/artifacts/credentials-for-write\032\004\010\002\020\000\020\003'
  _ARTIFACTCREDENTIALTYPE._serialized_start=979
  _ARTIFACTCREDENTIALTYPE._serialized_end=1094
  _ARTIFACTCREDENTIALINFO._serialized_start=109
  _ARTIFACTCREDENTIALINFO._serialized_end=374
  _ARTIFACTCREDENTIALINFO_HTTPHEADER._serialized_start=333
  _ARTIFACTCREDENTIALINFO_HTTPHEADER._serialized_end=374
  _GETCREDENTIALSFORREAD._serialized_start=377
  _GETCREDENTIALSFORREAD._serialized_end=675
  _GETCREDENTIALSFORREAD_RESPONSE._serialized_start=458
  _GETCREDENTIALSFORREAD_RESPONSE._serialized_end=578
  _GETCREDENTIALSFORWRITE._serialized_start=678
  _GETCREDENTIALSFORWRITE._serialized_end=977
  _GETCREDENTIALSFORWRITE_RESPONSE._serialized_start=458
  _GETCREDENTIALSFORWRITE_RESPONSE._serialized_end=578
  _DATABRICKSMLFLOWARTIFACTSSERVICE._serialized_start=1097
  _DATABRICKSMLFLOWARTIFACTSSERVICE._serialized_end=1537
DatabricksMlflowArtifactsService = service_reflection.GeneratedServiceType('DatabricksMlflowArtifactsService', (_service.Service,), dict(
  DESCRIPTOR = _DATABRICKSMLFLOWARTIFACTSSERVICE,
  __module__ = 'patched_databricks_artifacts_pb2'
  ))

DatabricksMlflowArtifactsService_Stub = service_reflection.GeneratedServiceStubType('DatabricksMlflowArtifactsService_Stub', (DatabricksMlflowArtifactsService,), dict(
  DESCRIPTOR = _DATABRICKSMLFLOWARTIFACTSSERVICE,
  __module__ = 'patched_databricks_artifacts_pb2'
  ))


# @@protoc_insertion_point(module_scope)



import json

from mlflow.protos.databricks_pb2 import (
    INTERNAL_ERROR,
    TEMPORARILY_UNAVAILABLE,
    ENDPOINT_NOT_FOUND,
    PERMISSION_DENIED,
    CUSTOMER_UNAUTHORIZED,
    REQUEST_LIMIT_EXCEEDED,
    BAD_REQUEST,
    INVALID_PARAMETER_VALUE,
    RESOURCE_DOES_NOT_EXIST,
    INVALID_STATE,
    RESOURCE_ALREADY_EXISTS,
    ErrorCode,
)

ERROR_CODE_TO_HTTP_STATUS = {
    ErrorCode.Name(INTERNAL_ERROR): 500,
    ErrorCode.Name(INVALID_STATE): 500,
    ErrorCode.Name(TEMPORARILY_UNAVAILABLE): 503,
    ErrorCode.Name(REQUEST_LIMIT_EXCEEDED): 429,
    ErrorCode.Name(ENDPOINT_NOT_FOUND): 404,
    ErrorCode.Name(RESOURCE_DOES_NOT_EXIST): 404,
    ErrorCode.Name(PERMISSION_DENIED): 403,
    ErrorCode.Name(CUSTOMER_UNAUTHORIZED): 401,
    ErrorCode.Name(BAD_REQUEST): 400,
    ErrorCode.Name(RESOURCE_ALREADY_EXISTS): 400,
    ErrorCode.Name(INVALID_PARAMETER_VALUE): 400,
}

HTTP_STATUS_TO_ERROR_CODE = dict((v, k) for k, v in ERROR_CODE_TO_HTTP_STATUS.items())
HTTP_STATUS_TO_ERROR_CODE[400] = ErrorCode.Name(BAD_REQUEST)
HTTP_STATUS_TO_ERROR_CODE[404] = ErrorCode.Name(ENDPOINT_NOT_FOUND)
HTTP_STATUS_TO_ERROR_CODE[500] = ErrorCode.Name(INTERNAL_ERROR)


def get_error_code(http_status):
    return ErrorCode.Value(
        HTTP_STATUS_TO_ERROR_CODE.get(http_status, ErrorCode.Name(INTERNAL_ERROR))
    )



import os
import time
import shutil

import pytest
import posixpath
from requests.models import Response
from unittest import mock
from unittest.mock import ANY

from mlflow.entities.file_info import FileInfo as FileInfoEntity
from mlflow.exceptions import MlflowException
from mlflow.protos.service_pb2 import ListArtifacts, FileInfo
from mlflow.store.artifact.artifact_repository_registry import get_artifact_repository

from mlflow_databricks_artifacts.protos.patched_databricks_artifacts_pb2 import (
    GetCredentialsForWrite,
    GetCredentialsForRead,
    ArtifactCredentialType,
    ArtifactCredentialInfo,
)
from mlflow_databricks_artifacts.store.artifact_repo import (
    DatabricksArtifactRepository,
    _MAX_CREDENTIALS_REQUEST_SIZE,
)

DATABRICKS_ARTIFACT_REPOSITORY_PACKAGE = (
    "mlflow_databricks_artifacts.store.artifact_repo"
)
DATABRICKS_ARTIFACT_REPOSITORY = (
    DATABRICKS_ARTIFACT_REPOSITORY_PACKAGE + ".DatabricksArtifactRepository"
)

MOCK_AZURE_SIGNED_URI = "http://this_is_a_mock_sas_for_azure"
MOCK_ADLS_GEN2_SIGNED_URI = "http://this_is_a_mock_sas_for_adls_gen2"
MOCK_AWS_SIGNED_URI = "http://this_is_a_mock_presigned_uri_for_aws?"
MOCK_GCP_SIGNED_URL = "http://this_is_a_mock_signed_url_for_gcp?"
MOCK_RUN_ID = "MOCK-RUN-ID"
MOCK_HEADERS = [
    ArtifactCredentialInfo.HttpHeader(name="Mock-Name1", value="Mock-Value1"),
    ArtifactCredentialInfo.HttpHeader(name="Mock-Name2", value="Mock-Value2"),
]
MOCK_RUN_ROOT_URI = "dbfs:/databricks/mlflow-tracking/MOCK-EXP/MOCK-RUN-ID/artifacts"
MOCK_SUBDIR = "subdir/path"
MOCK_SUBDIR_ROOT_URI = posixpath.join(MOCK_RUN_ROOT_URI, MOCK_SUBDIR)


@pytest.fixture()
def databricks_artifact_repo():
    with mock.patch(
        DATABRICKS_ARTIFACT_REPOSITORY + "._get_run_artifact_root"
    ) as get_run_artifact_root_mock:
        get_run_artifact_root_mock.return_value = MOCK_RUN_ROOT_URI
        return get_artifact_repository(
            "dbfs:/databricks/mlflow-tracking/MOCK-EXP/MOCK-RUN-ID/artifacts"
        )


@pytest.fixture()
def test_file(tmpdir):
    test_file_content = "Hello 🍆🍔".encode("utf-8")
    p = tmpdir.join("test.txt")
    with open(p.strpath, "wb") as f:
        f.write(test_file_content)
    return p


@pytest.fixture()
def test_dir(tmpdir):
    test_file_content = "World 🍆🍔🍆".encode("utf-8")
    with open(tmpdir.mkdir("subdir").join("test.txt").strpath, "wb") as f:
        f.write(test_file_content)
    with open(tmpdir.join("test.txt").strpath, "wb") as f:
        f.write(bytes(test_file_content))
    with open(tmpdir.join("empty-file.txt").strpath, "w"):
        pass
    return tmpdir


class TestDatabricksArtifactRepository(object):
    def test_init_validation_and_cleaning(self):
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_run_artifact_root"
        ) as get_run_artifact_root_mock:
            get_run_artifact_root_mock.return_value = MOCK_RUN_ROOT_URI
            # Basic artifact uri
            repo = get_artifact_repository(
                "dbfs:/databricks/mlflow-tracking/MOCK-EXP/MOCK-RUN-ID/artifacts"
            )
            assert (
                repo.artifact_uri == "dbfs:/databricks/mlflow-tracking/"
                "MOCK-EXP/MOCK-RUN-ID/artifacts"
            )
            assert repo.run_id == MOCK_RUN_ID
            assert repo.run_relative_artifact_repo_root_path == ""

            with pytest.raises(
                MlflowException, match="DBFS URI must be of the form dbfs"
            ):
                DatabricksArtifactRepository("s3://test")
            with pytest.raises(MlflowException, match="Artifact URI incorrect"):
                DatabricksArtifactRepository("dbfs:/databricks/mlflow/EXP/RUN/artifact")
            with pytest.raises(
                MlflowException, match="DBFS URI must be of the form dbfs"
            ):
                DatabricksArtifactRepository(
                    "dbfs://scope:key@notdatabricks/databricks/mlflow-tracking/experiment/1/run/2"
                )

    @pytest.mark.parametrize(
        ("artifact_uri", "expected_uri", "expected_db_uri"),
        [
            (
                "dbfs:/databricks/mlflow-tracking/experiment/1/run/2",
                "dbfs:/databricks/mlflow-tracking/experiment/1/run/2",
                "databricks://getTrackingUriDefault",
            ),  # see test body for the mock
            (
                "dbfs://@databricks/databricks/mlflow-tracking/experiment/1/run/2",
                "dbfs:/databricks/mlflow-tracking/experiment/1/run/2",
                "databricks",
            ),
            (
                "dbfs://someProfile@databricks/databricks/mlflow-tracking/experiment/1/run/2",
                "dbfs:/databricks/mlflow-tracking/experiment/1/run/2",
                "databricks://someProfile",
            ),
            (
                "dbfs://scope:key@databricks/databricks/mlflow-tracking/experiment/1/run/2",
                "dbfs:/databricks/mlflow-tracking/experiment/1/run/2",
                "databricks://scope:key",
            ),
            (
                "dbfs:/databricks/mlflow-tracking/MOCK-EXP/MOCK-RUN-ID/artifacts",
                "dbfs:/databricks/mlflow-tracking/MOCK-EXP/MOCK-RUN-ID/artifacts",
                "databricks://getTrackingUriDefault",
            ),
            (
                "dbfs:/databricks/mlflow-tracking/MOCK-EXP/MOCK-RUN-ID/awesome/path",
                "dbfs:/databricks/mlflow-tracking/MOCK-EXP/MOCK-RUN-ID/awesome/path",
                "databricks://getTrackingUriDefault",
            ),
        ],
    )
    def test_init_artifact_uri(self, artifact_uri, expected_uri, expected_db_uri):
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY_PACKAGE + ".get_databricks_host_creds",
            return_value=None,
        ), mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_run_artifact_root",
            return_value="whatever",
        ), mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY_PACKAGE + ".get_tracking_uri",
            return_value="databricks://getTrackingUriDefault",
        ):
            repo = DatabricksArtifactRepository(artifact_uri)
            assert repo.artifact_uri == expected_uri
            assert repo.databricks_profile_uri == expected_db_uri

    @pytest.mark.parametrize(
        ("artifact_uri", "expected_relative_path"),
        [
            ("dbfs:/databricks/mlflow-tracking/MOCK-EXP/MOCK-RUN-ID/artifacts", ""),
            (
                "dbfs:/databricks/mlflow-tracking/MOCK-EXP/MOCK-RUN-ID/artifacts/arty",
                "arty",
            ),
            (
                "dbfs://prof@databricks/databricks/mlflow-tracking/MOCK-EXP/MOCK-RUN-ID/artifacts/arty",  # pylint: disable=line-too-long
                "arty",
            ),
            (
                "dbfs:/databricks/mlflow-tracking/MOCK-EXP/MOCK-RUN-ID/awesome/path",
                "../awesome/path",
            ),
        ],
    )
    def test_run_relative_artifact_repo_root_path(
        self, artifact_uri, expected_relative_path
    ):
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_run_artifact_root"
        ) as get_run_artifact_root_mock:
            get_run_artifact_root_mock.return_value = MOCK_RUN_ROOT_URI
            # Basic artifact uri
            repo = get_artifact_repository(artifact_uri)
            assert repo.run_id == MOCK_RUN_ID
            assert repo.run_relative_artifact_repo_root_path == expected_relative_path

    def test_extract_run_id(self):
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_run_artifact_root"
        ) as get_run_artifact_root_mock:
            get_run_artifact_root_mock.return_value = MOCK_RUN_ROOT_URI
            expected_run_id = "RUN_ID"
            repo = get_artifact_repository(
                "dbfs:/databricks/mlflow-tracking/EXP/RUN_ID/artifact"
            )
            assert repo.run_id == expected_run_id
            repo = get_artifact_repository(
                "dbfs:/databricks/mlflow-tracking/EXP_ID/RUN_ID/artifacts"
            )
            assert repo.run_id == expected_run_id
            repo = get_artifact_repository(
                "dbfs:/databricks///mlflow-tracking///EXP_ID///RUN_ID///artifacts/"
            )
            assert repo.run_id == expected_run_id
            repo = get_artifact_repository(
                "dbfs:/databricks///mlflow-tracking//EXP_ID//RUN_ID///artifacts//"
            )
            assert repo.run_id == expected_run_id

    @pytest.mark.parametrize(
        ("artifact_path", "expected_location"),
        [(None, "test.txt"), ("output", "output/test.txt"), ("", "test.txt")],
    )
    def test_log_artifact_azure(
        self, databricks_artifact_repo, test_file, artifact_path, expected_location
    ):
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credential_infos_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._azure_upload_file"
        ) as azure_upload_mock:
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_AZURE_SIGNED_URI,
                type=ArtifactCredentialType.AZURE_SAS_URI,
            )
            write_credential_infos_mock.return_value = [mock_credential_info]
            azure_upload_mock.return_value = None
            databricks_artifact_repo.log_artifact(test_file.strpath, artifact_path)
            write_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=[expected_location]
            )
            azure_upload_mock.assert_called_with(
                mock_credential_info, test_file.strpath, expected_location
            )

    @pytest.mark.parametrize(
        ("artifact_path", "expected_location"), [(None, "test.txt")]
    )
    def test_log_artifact_azure_with_headers(
        self, databricks_artifact_repo, test_file, artifact_path, expected_location
    ):
        mock_azure_headers = {
            "x-ms-encryption-scope": "test-scope",
            "x-ms-tags": "some-tags",
            "x-ms-blob-type": "some-type",
        }
        filtered_azure_headers = {
            "x-ms-encryption-scope": "test-scope",
            "x-ms-tags": "some-tags",
        }
        mock_response = Response()
        mock_response.status_code = 200
        mock_response.close = lambda: None
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credential_infos_mock, mock.patch(
            "mlflow_databricks_artifacts.utils.rest_utils.cloud_storage_http_request"
        ) as request_mock:
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_AZURE_SIGNED_URI,
                type=ArtifactCredentialType.AZURE_SAS_URI,
                headers=[
                    ArtifactCredentialInfo.HttpHeader(
                        name=header_name, value=header_value
                    )
                    for header_name, header_value in mock_azure_headers.items()
                ],
            )
            write_credential_infos_mock.return_value = [mock_credential_info]
            request_mock.return_value = mock_response
            databricks_artifact_repo.log_artifact(test_file.strpath, artifact_path)
            write_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=[expected_location]
            )
            request_mock.assert_called_with(
                "put",
                MOCK_AZURE_SIGNED_URI + "?comp=blocklist",
                data=ANY,
                headers=filtered_azure_headers,
            )

    def test_log_artifact_azure_blob_client_sas_error(
        self, databricks_artifact_repo, test_file
    ):
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credential_infos_mock, mock.patch(
            "azure.storage.blob.BlobClient.from_blob_url"
        ) as mock_create_blob_client:
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_AZURE_SIGNED_URI,
                type=ArtifactCredentialType.AZURE_SAS_URI,
            )
            write_credential_infos_mock.return_value = [mock_credential_info]
            mock_create_blob_client.side_effect = MlflowException("MOCK ERROR")
            with pytest.raises(MlflowException, match=r".+"):
                databricks_artifact_repo.log_artifact(test_file.strpath)
            write_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=ANY
            )

    @pytest.mark.parametrize(
        ("artifact_path", "expected_location"),
        [(None, "test.txt"), ("output", "output/test.txt"), ("", "test.txt")],
    )
    def test_log_artifact_adls_gen2(
            self, databricks_artifact_repo, test_file, artifact_path, expected_location
    ):
        with mock.patch(
                DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credential_infos_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._azure_adls_gen2_upload_file"
        ) as azure_adls_gen2_upload_mock:
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_ADLS_GEN2_SIGNED_URI,
                type=ArtifactCredentialType.AZURE_ADLS_GEN2_SAS_URI,
            )
            write_credential_infos_mock.return_value = [mock_credential_info]
            azure_adls_gen2_upload_mock.return_value = None
            databricks_artifact_repo.log_artifact(test_file.strpath, artifact_path)
            write_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=[expected_location]
            )
            azure_adls_gen2_upload_mock.assert_called_with(
                mock_credential_info, test_file.strpath, expected_location
            )

    @pytest.mark.parametrize(("artifact_path", "expected_location"), [(None, "test.txt")])
    def test_log_artifact_adls_gen2_with_headers(
            self, databricks_artifact_repo, test_file, artifact_path, expected_location
    ):
        mock_azure_headers = {
            "x-ms-content-type": "test-type",
            "x-ms-owner": "some-owner",
            "x-ms-something_not_supported": "some-value",
        }
        filtered_azure_headers = {
            "x-ms-content-type": "test-type",
            "x-ms-owner": "some-owner",
        }
        mock_response = Response()
        mock_response.status_code = 200
        mock_response.close = lambda: None
        with mock.patch(
                DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credential_infos_mock, mock.patch(
            "mlflow_databricks_artifacts.utils.rest_utils.cloud_storage_http_request"
        ) as request_mock, mock.patch(
            "mlflow_databricks_artifacts.store.artifact_repo._AZURE_MAX_BLOCK_CHUNK_SIZE",
            5,
        ):
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_ADLS_GEN2_SIGNED_URI,
                type=ArtifactCredentialType.AZURE_ADLS_GEN2_SAS_URI,
                headers=[
                    ArtifactCredentialInfo.HttpHeader(name=header_name, value=header_value)
                    for header_name, header_value in mock_azure_headers.items()
                ],
            )
            write_credential_infos_mock.return_value = [mock_credential_info]
            request_mock.return_value = mock_response
            databricks_artifact_repo.log_artifact(test_file.strpath, artifact_path)
            write_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=[expected_location]
            )
            # test with block size 5
            request_mock.assert_any_call(
                "put",
                MOCK_ADLS_GEN2_SIGNED_URI + "?resource=file",
                headers=filtered_azure_headers,
                )
            request_mock.assert_any_call(
                "patch",
                MOCK_ADLS_GEN2_SIGNED_URI + "?action=append&position=0",
                data=ANY,
                headers=filtered_azure_headers,
                )
            request_mock.assert_any_call(
                "patch",
                MOCK_ADLS_GEN2_SIGNED_URI + "?action=append&position=5",
                data=ANY,
                headers=filtered_azure_headers,
                )
            request_mock.assert_any_call(
                "patch",
                MOCK_ADLS_GEN2_SIGNED_URI + "?action=append&position=10",
                data=ANY,
                headers=filtered_azure_headers,
                )
            request_mock.assert_called_with(
                "patch",
                MOCK_ADLS_GEN2_SIGNED_URI + "?action=flush&position=14",
                headers=filtered_azure_headers,
                )

    def test_log_artifact_adls_gen2_flush_error(self, databricks_artifact_repo, test_file):
        mock_successful_response = Response()
        mock_successful_response.status_code = 200
        mock_successful_response.close = lambda: None
        mock_error_response = MlflowException("MOCK ERROR")
        mock_credential_info = ArtifactCredentialInfo(
            signed_uri=MOCK_ADLS_GEN2_SIGNED_URI,
            type=ArtifactCredentialType.AZURE_ADLS_GEN2_SAS_URI,
        )
        with mock.patch(
                DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos",
                return_value=[mock_credential_info],
        ) as write_credential_infos_mock, mock.patch(
            "mlflow_databricks_artifacts.utils.rest_utils.cloud_storage_http_request",
            side_effect=[mock_successful_response, mock_error_response],
        ) as request_mock:
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_ADLS_GEN2_SIGNED_URI,
                type=ArtifactCredentialType.AZURE_ADLS_GEN2_SAS_URI,
            )
            with pytest.raises(MlflowException, match=r"MOCK ERROR"):
                databricks_artifact_repo.log_artifact(test_file.strpath)
            write_credential_infos_mock.assert_called_with(run_id=MOCK_RUN_ID, paths=ANY)
            assert request_mock.mock_calls == [
                mock.call(
                    "put",
                    MOCK_ADLS_GEN2_SIGNED_URI + "?resource=file",
                    headers={},
                    ),
                mock.call(
                    "patch",
                    MOCK_ADLS_GEN2_SIGNED_URI + "?action=append&position=0&flush=true",
                    data=ANY,
                    headers={},
                    ),
            ]

    @pytest.mark.parametrize(
        ("artifact_path", "expected_location"), [(None, "test.txt")]
    )
    def test_log_artifact_aws(
        self, databricks_artifact_repo, test_file, artifact_path, expected_location
    ):
        mock_response = Response()
        mock_response.status_code = 200
        mock_response.close = lambda: None
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credential_infos_mock, mock.patch(
            "mlflow_databricks_artifacts.utils.rest_utils.cloud_storage_http_request"
        ) as request_mock:
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_AWS_SIGNED_URI,
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            )
            write_credential_infos_mock.return_value = [mock_credential_info]
            request_mock.return_value = mock_response
            databricks_artifact_repo.log_artifact(test_file.strpath, artifact_path)
            write_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=[expected_location]
            )
            request_mock.assert_called_with(
                "put", MOCK_AWS_SIGNED_URI, data=ANY, headers={}
            )

    @pytest.mark.parametrize(
        ("artifact_path", "expected_location"), [(None, "test.txt")]
    )
    def test_log_artifact_aws_with_headers(
        self, databricks_artifact_repo, test_file, artifact_path, expected_location
    ):
        expected_headers = {header.name: header.value for header in MOCK_HEADERS}
        mock_response = Response()
        mock_response.status_code = 200
        mock_response.close = lambda: None
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credential_infos_mock, mock.patch(
            "mlflow_databricks_artifacts.utils.rest_utils.cloud_storage_http_request"
        ) as request_mock:
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_AWS_SIGNED_URI,
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
                headers=MOCK_HEADERS,
            )
            write_credential_infos_mock.return_value = [mock_credential_info]
            request_mock.return_value = mock_response
            databricks_artifact_repo.log_artifact(test_file.strpath, artifact_path)
            write_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=[expected_location]
            )
            request_mock.assert_called_with(
                "put", MOCK_AWS_SIGNED_URI, data=ANY, headers=expected_headers
            )

    def test_log_artifact_aws_presigned_url_error(
        self, databricks_artifact_repo, test_file
    ):
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credential_infos_mock, mock.patch(
            "mlflow_databricks_artifacts.utils.rest_utils.cloud_storage_http_request"
        ) as request_mock:
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_AWS_SIGNED_URI,
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            )
            write_credential_infos_mock.return_value = [mock_credential_info]
            request_mock.side_effect = MlflowException("MOCK ERROR")
            with pytest.raises(MlflowException, match="MOCK ERROR"):
                databricks_artifact_repo.log_artifact(test_file.strpath)
            write_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=ANY
            )

    @pytest.mark.parametrize(
        ("artifact_path", "expected_location"), [(None, "test.txt")]
    )
    def test_log_artifact_gcp(
        self, databricks_artifact_repo, test_file, artifact_path, expected_location
    ):
        mock_response = Response()
        mock_response.status_code = 200
        mock_response.close = lambda: None
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credential_infos_mock, mock.patch(
            "mlflow_databricks_artifacts.utils.rest_utils.cloud_storage_http_request"
        ) as request_mock:
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_GCP_SIGNED_URL,
                type=ArtifactCredentialType.GCP_SIGNED_URL,
            )
            write_credential_infos_mock.return_value = [mock_credential_info]
            request_mock.return_value = mock_response
            databricks_artifact_repo.log_artifact(test_file.strpath, artifact_path)
            write_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=[expected_location]
            )
            request_mock.assert_called_with(
                "put", MOCK_GCP_SIGNED_URL, data=ANY, headers={}
            )

    @pytest.mark.parametrize(
        ("artifact_path", "expected_location"), [(None, "test.txt")]
    )
    def test_log_artifact_gcp_with_headers(
        self, databricks_artifact_repo, test_file, artifact_path, expected_location
    ):
        expected_headers = {header.name: header.value for header in MOCK_HEADERS}
        mock_response = Response()
        mock_response.status_code = 200
        mock_response.close = lambda: None
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credential_infos_mock, mock.patch(
            "mlflow_databricks_artifacts.utils.rest_utils.cloud_storage_http_request"
        ) as request_mock:
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_GCP_SIGNED_URL,
                type=ArtifactCredentialType.GCP_SIGNED_URL,
                headers=MOCK_HEADERS,
            )
            write_credential_infos_mock.return_value = [mock_credential_info]
            request_mock.return_value = mock_response
            databricks_artifact_repo.log_artifact(test_file.strpath, artifact_path)
            write_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=[expected_location]
            )
            request_mock.assert_called_with(
                "put", MOCK_GCP_SIGNED_URL, data=ANY, headers=expected_headers
            )

    def test_log_artifact_gcp_presigned_url_error(
        self, databricks_artifact_repo, test_file
    ):
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credential_infos_mock, mock.patch(
            "mlflow_databricks_artifacts.utils.rest_utils.cloud_storage_http_request"
        ) as request_mock:
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_GCP_SIGNED_URL,
                type=ArtifactCredentialType.GCP_SIGNED_URL,
            )
            write_credential_infos_mock.return_value = [mock_credential_info]
            request_mock.side_effect = MlflowException("MOCK ERROR")
            with pytest.raises(MlflowException, match="MOCK ERROR"):
                databricks_artifact_repo.log_artifact(test_file.strpath)
            write_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=ANY
            )

    @pytest.mark.parametrize(
        ("artifact_path", "expected_location"),
        [
            (None, posixpath.join(MOCK_SUBDIR, "test.txt")),
            ("test_path", posixpath.join(MOCK_SUBDIR, "test_path/test.txt")),
        ],
    )
    def test_log_artifact_with_relative_path(
        self, test_file, artifact_path, expected_location
    ):
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_run_artifact_root"
        ) as get_run_artifact_root_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credential_infos_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._upload_to_cloud"
        ) as upload_mock:
            get_run_artifact_root_mock.return_value = MOCK_RUN_ROOT_URI
            databricks_artifact_repo = get_artifact_repository(MOCK_SUBDIR_ROOT_URI)
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_AZURE_SIGNED_URI,
                type=ArtifactCredentialType.AZURE_SAS_URI,
            )
            write_credential_infos_mock.return_value = [mock_credential_info]
            upload_mock.return_value = None
            databricks_artifact_repo.log_artifact(test_file.strpath, artifact_path)
            write_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=[expected_location]
            )
            upload_mock.assert_called_with(
                cloud_credential_info=mock_credential_info,
                src_file_path=test_file.strpath,
                dst_run_relative_artifact_path=expected_location,
            )

    def test_list_artifacts(self, databricks_artifact_repo):
        list_artifact_file_proto_mock = [
            FileInfo(path="a.txt", is_dir=False, file_size=0)
        ]
        list_artifacts_dir_proto_mock = [
            FileInfo(path="test/a.txt", is_dir=False, file_size=100),
            FileInfo(path="test/dir", is_dir=True, file_size=0),
        ]
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._call_endpoint"
        ) as call_endpoint_mock:
            list_artifact_response_proto = ListArtifacts.Response(
                root_uri="", files=list_artifacts_dir_proto_mock, next_page_token=None
            )
            call_endpoint_mock.return_value = list_artifact_response_proto
            artifacts = databricks_artifact_repo.list_artifacts("test/")
            assert isinstance(artifacts, list)
            assert isinstance(artifacts[0], FileInfoEntity)
            assert len(artifacts) == 2
            assert artifacts[0].path == "test/a.txt"
            assert artifacts[0].is_dir is False
            assert artifacts[0].file_size == 100
            assert artifacts[1].path == "test/dir"
            assert artifacts[1].is_dir is True
            assert artifacts[1].file_size is None

            # Calling list_artifacts() on a path that's a file should return an empty list
            list_artifact_response_proto = ListArtifacts.Response(
                root_uri="", files=list_artifact_file_proto_mock
            )
            call_endpoint_mock.return_value = list_artifact_response_proto
            artifacts = databricks_artifact_repo.list_artifacts("a.txt")
            assert len(artifacts) == 0

    def test_list_artifacts_with_relative_path(self):
        list_artifact_file_proto_mock = [
            FileInfo(
                path=posixpath.join(MOCK_SUBDIR, "a.txt"), is_dir=False, file_size=0
            )
        ]
        list_artifacts_dir_proto_mock = [
            FileInfo(
                path=posixpath.join(MOCK_SUBDIR, "test/a.txt"),
                is_dir=False,
                file_size=100,
            ),
            FileInfo(
                path=posixpath.join(MOCK_SUBDIR, "test/dir"), is_dir=True, file_size=0
            ),
        ]
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_run_artifact_root"
        ) as get_run_artifact_root_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY_PACKAGE + ".message_to_json"
        ) as message_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._call_endpoint"
        ) as call_endpoint_mock:
            get_run_artifact_root_mock.return_value = MOCK_RUN_ROOT_URI
            list_artifact_response_proto = ListArtifacts.Response(
                root_uri="", files=list_artifacts_dir_proto_mock, next_page_token=None
            )
            call_endpoint_mock.return_value = list_artifact_response_proto
            message_mock.return_value = None
            databricks_artifact_repo = get_artifact_repository(MOCK_SUBDIR_ROOT_URI)
            artifacts = databricks_artifact_repo.list_artifacts("test")
            assert isinstance(artifacts, list)
            assert isinstance(artifacts[0], FileInfoEntity)
            assert len(artifacts) == 2
            assert artifacts[0].path == "test/a.txt"
            assert artifacts[0].is_dir is False
            assert artifacts[0].file_size == 100
            assert artifacts[1].path == "test/dir"
            assert artifacts[1].is_dir is True
            assert artifacts[1].file_size is None
            message_mock.assert_called_with(
                ListArtifacts(
                    run_id=MOCK_RUN_ID, path=posixpath.join(MOCK_SUBDIR, "test")
                )
            )

            # Calling list_artifacts() on a relative path that's a file should return an empty list
            list_artifact_response_proto = ListArtifacts.Response(
                root_uri="", files=list_artifact_file_proto_mock, next_page_token=None
            )
            call_endpoint_mock.return_value = list_artifact_response_proto
            artifacts = databricks_artifact_repo.list_artifacts("a.txt")
            assert len(artifacts) == 0

    def test_list_artifacts_handles_pagination(self, databricks_artifact_repo):
        list_artifacts_proto_mock_1 = [
            FileInfo(path="a.txt", is_dir=False, file_size=100),
            FileInfo(path="b", is_dir=True, file_size=0),
        ]
        list_artifacts_proto_mock_2 = [
            FileInfo(path="c.txt", is_dir=False, file_size=100),
            FileInfo(path="d", is_dir=True, file_size=0),
        ]
        list_artifacts_proto_mock_3 = [
            FileInfo(path="e.txt", is_dir=False, file_size=100),
            FileInfo(path="f", is_dir=True, file_size=0),
        ]
        list_artifacts_proto_mock_4 = []
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY_PACKAGE + ".message_to_json"
        ) as message_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._call_endpoint"
        ) as call_endpoint_mock:
            list_artifact_paginated_response_protos = [
                ListArtifacts.Response(
                    root_uri="", files=list_artifacts_proto_mock_1, next_page_token="2"
                ),
                ListArtifacts.Response(
                    root_uri="", files=list_artifacts_proto_mock_2, next_page_token="4"
                ),
                ListArtifacts.Response(
                    root_uri="", files=list_artifacts_proto_mock_3, next_page_token="6"
                ),
                ListArtifacts.Response(
                    root_uri="", files=list_artifacts_proto_mock_4, next_page_token="8"
                ),
            ]
            call_endpoint_mock.side_effect = list_artifact_paginated_response_protos
            message_mock.return_value = None
            artifacts = databricks_artifact_repo.list_artifacts()
            assert {"a.txt", "b", "c.txt", "d", "e.txt", "f"} == {
                file.path for file in artifacts
            }
            calls = [
                mock.call(ListArtifacts(run_id=MOCK_RUN_ID, path="")),
                mock.call(ListArtifacts(run_id=MOCK_RUN_ID, path="", page_token="2")),
                mock.call(ListArtifacts(run_id=MOCK_RUN_ID, path="", page_token="4")),
                mock.call(ListArtifacts(run_id=MOCK_RUN_ID, path="", page_token="6")),
            ]
            message_mock.assert_has_calls(calls)

    def test_get_read_credential_infos_handles_pagination(
        self, databricks_artifact_repo
    ):
        """
        Verifies that the `get_read_credential_infos` method, which is used to resolve read access
        credentials for a collection of artifacts, handles paginated responses properly, issuing
        incremental requests until all pages have been consumed
        """
        credential_infos_mock_1 = [
            ArtifactCredentialInfo(
                signed_uri="http://mock_url_1",
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            ),
            ArtifactCredentialInfo(
                signed_uri="http://mock_url_2",
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            ),
        ]
        credential_infos_mock_2 = [
            ArtifactCredentialInfo(
                signed_uri="http://mock_url_3",
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            )
        ]
        credential_infos_mock_3 = []

        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY_PACKAGE + ".message_to_json"
        ) as message_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._call_endpoint"
        ) as call_endpoint_mock:
            get_credentials_for_read_responses = [
                GetCredentialsForRead.Response(
                    credential_infos=credential_infos_mock_1, next_page_token="2"
                ),
                GetCredentialsForRead.Response(
                    credential_infos=credential_infos_mock_2, next_page_token="3"
                ),
                GetCredentialsForRead.Response(
                    credential_infos=credential_infos_mock_3
                ),
            ]
            call_endpoint_mock.side_effect = get_credentials_for_read_responses
            read_credential_infos = databricks_artifact_repo._get_read_credential_infos(
                MOCK_RUN_ID,
                ["testpath"],
            )
            assert (
                read_credential_infos
                == credential_infos_mock_1 + credential_infos_mock_2
            )
            message_mock.assert_has_calls(
                [
                    mock.call(
                        GetCredentialsForRead(run_id=MOCK_RUN_ID, path=["testpath"])
                    ),
                    mock.call(
                        GetCredentialsForRead(
                            run_id=MOCK_RUN_ID, path=["testpath"], page_token="2"
                        )
                    ),
                    mock.call(
                        GetCredentialsForRead(
                            run_id=MOCK_RUN_ID, path=["testpath"], page_token="3"
                        )
                    ),
                ]
            )
            assert call_endpoint_mock.call_count == 3

    def test_get_read_credential_infos_respects_max_request_size(
        self, databricks_artifact_repo
    ):
        """
        Verifies that the `_get_read_credential_infos` method, which is used to resolve read access
        credentials for a collection of artifacts, handles paginated responses properly, issuing
        incremental requests until all pages have been consumed
        """
        assert _MAX_CREDENTIALS_REQUEST_SIZE == 2000, (
            "The maximum request size configured by the client should be consistent with the"
            " Databricks backend. Only update this value of the backend limit has changed."
        )

        # Create 3 chunks of paths, two of which have the maximum request size and one of which
        # is smaller than the maximum chunk size. Aggregate and pass these to
        # `_get_read_credential_infos`, validating that this method decomposes the aggregate
        # list into these expected chunks and makes 3 separate requests
        paths_chunk_1 = ["path1"] * _MAX_CREDENTIALS_REQUEST_SIZE
        paths_chunk_2 = ["path2"] * _MAX_CREDENTIALS_REQUEST_SIZE
        paths_chunk_3 = ["path3"] * 5
        credential_infos_mock_1 = [
            ArtifactCredentialInfo(
                signed_uri="http://mock_url_1",
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            )
            for _ in range(_MAX_CREDENTIALS_REQUEST_SIZE)
        ]
        credential_infos_mock_2 = [
            ArtifactCredentialInfo(
                signed_uri="http://mock_url_2",
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            )
            for _ in range(_MAX_CREDENTIALS_REQUEST_SIZE)
        ]
        credential_infos_mock_3 = [
            ArtifactCredentialInfo(
                signed_uri="http://mock_url_3",
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            )
            for _ in range(5)
        ]

        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY_PACKAGE + ".message_to_json"
        ) as message_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._call_endpoint"
        ) as call_endpoint_mock:
            call_endpoint_mock.side_effect = [
                GetCredentialsForRead.Response(
                    credential_infos=credential_infos_mock_1
                ),
                GetCredentialsForRead.Response(
                    credential_infos=credential_infos_mock_2
                ),
                GetCredentialsForRead.Response(
                    credential_infos=credential_infos_mock_3
                ),
            ]

            databricks_artifact_repo._get_read_credential_infos(
                MOCK_RUN_ID,
                paths_chunk_1 + paths_chunk_2 + paths_chunk_3,
            )
            assert call_endpoint_mock.call_count == 3
            assert message_mock.call_count == 3
            message_mock.assert_has_calls(
                [
                    mock.call(
                        GetCredentialsForRead(run_id=MOCK_RUN_ID, path=paths_chunk_1)
                    ),
                    mock.call(
                        GetCredentialsForRead(run_id=MOCK_RUN_ID, path=paths_chunk_2)
                    ),
                    mock.call(
                        GetCredentialsForRead(run_id=MOCK_RUN_ID, path=paths_chunk_3)
                    ),
                ]
            )

    def test_get_write_credential_infos_handles_pagination(
        self, databricks_artifact_repo
    ):
        """
        Verifies that the `_get_write_credential_infos` method, which is used to resolve write
        access credentials for a collection of artifacts, handles paginated responses properly,
        issuing incremental requests until all pages have been consumed
        """
        credential_infos_mock_1 = [
            ArtifactCredentialInfo(
                signed_uri="http://mock_url_1",
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            ),
            ArtifactCredentialInfo(
                signed_uri="http://mock_url_2",
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            ),
        ]
        credential_infos_mock_2 = [
            ArtifactCredentialInfo(
                signed_uri="http://mock_url_3",
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            )
        ]
        credential_infos_mock_3 = []

        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY_PACKAGE + ".message_to_json"
        ) as message_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._call_endpoint"
        ) as call_endpoint_mock:
            get_credentials_for_write_responses = [
                GetCredentialsForWrite.Response(
                    credential_infos=credential_infos_mock_1, next_page_token="2"
                ),
                GetCredentialsForWrite.Response(
                    credential_infos=credential_infos_mock_2, next_page_token="3"
                ),
                GetCredentialsForWrite.Response(
                    credential_infos=credential_infos_mock_3
                ),
            ]
            call_endpoint_mock.side_effect = get_credentials_for_write_responses
            write_credential_infos = (
                databricks_artifact_repo._get_write_credential_infos(
                    MOCK_RUN_ID,
                    ["testpath"],
                )
            )
            assert (
                write_credential_infos
                == credential_infos_mock_1 + credential_infos_mock_2
            )
            message_mock.assert_has_calls(
                [
                    mock.call(
                        GetCredentialsForWrite(run_id=MOCK_RUN_ID, path=["testpath"])
                    ),
                    mock.call(
                        GetCredentialsForWrite(
                            run_id=MOCK_RUN_ID, path=["testpath"], page_token="2"
                        )
                    ),
                    mock.call(
                        GetCredentialsForWrite(
                            run_id=MOCK_RUN_ID, path=["testpath"], page_token="3"
                        )
                    ),
                ]
            )
            assert call_endpoint_mock.call_count == 3

    def test_get_write_credential_infos_respects_max_request_size(
        self, databricks_artifact_repo
    ):
        """
        Verifies that the `_get_write_credential_infos` method, which is used to resolve write
        access credentials for a collection of artifacts, batches requests according to a maximum
        request size configured by the backend
        """
        # Create 3 chunks of paths, two of which have the maximum request size and one of which
        # is smaller than the maximum chunk size. Aggregate and pass these to
        # `_get_write_credential_infos`, validating that this method decomposes the aggregate
        # list into these expected chunks and makes 3 separate requests
        paths_chunk_1 = ["path1"] * 2000
        paths_chunk_2 = ["path2"] * 2000
        paths_chunk_3 = ["path3"] * 5
        credential_infos_mock_1 = [
            ArtifactCredentialInfo(
                signed_uri="http://mock_url_1",
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            )
            for _ in range(2000)
        ]
        credential_infos_mock_2 = [
            ArtifactCredentialInfo(
                signed_uri="http://mock_url_2",
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            )
            for _ in range(2000)
        ]
        credential_infos_mock_3 = [
            ArtifactCredentialInfo(
                signed_uri="http://mock_url_3",
                type=ArtifactCredentialType.AWS_PRESIGNED_URL,
            )
            for _ in range(5)
        ]

        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY_PACKAGE + ".message_to_json"
        ) as message_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._call_endpoint"
        ) as call_endpoint_mock:
            call_endpoint_mock.side_effect = [
                GetCredentialsForWrite.Response(
                    credential_infos=credential_infos_mock_1
                ),
                GetCredentialsForWrite.Response(
                    credential_infos=credential_infos_mock_2
                ),
                GetCredentialsForWrite.Response(
                    credential_infos=credential_infos_mock_3
                ),
            ]

            databricks_artifact_repo._get_write_credential_infos(
                MOCK_RUN_ID,
                paths_chunk_1 + paths_chunk_2 + paths_chunk_3,
            )
            assert call_endpoint_mock.call_count == message_mock.call_count == 3
            message_mock.assert_has_calls(
                [
                    mock.call(
                        GetCredentialsForWrite(run_id=MOCK_RUN_ID, path=paths_chunk_1)
                    ),
                    mock.call(
                        GetCredentialsForWrite(run_id=MOCK_RUN_ID, path=paths_chunk_2)
                    ),
                    mock.call(
                        GetCredentialsForWrite(run_id=MOCK_RUN_ID, path=paths_chunk_3)
                    ),
                ]
            )

    @pytest.mark.parametrize(
        ("remote_file_path", "local_path", "cloud_credential_type"),
        [
            ("test_file.txt", "", ArtifactCredentialType.AZURE_SAS_URI),
            ("test_file.txt", None, ArtifactCredentialType.AZURE_SAS_URI),
            ("output/test_file", None, ArtifactCredentialType.AZURE_SAS_URI),
            ("test_file.txt", "", ArtifactCredentialType.AWS_PRESIGNED_URL),
            ("test_file.txt", "", ArtifactCredentialType.GCP_SIGNED_URL),
        ],
    )
    def test_databricks_download_file(
        self,
        databricks_artifact_repo,
        remote_file_path,
        local_path,
        cloud_credential_type,
    ):
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_read_credential_infos"
        ) as read_credential_infos_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + ".list_artifacts"
        ) as get_list_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._download_from_cloud"
        ) as download_mock:
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_AZURE_SIGNED_URI, type=cloud_credential_type
            )
            read_credential_infos_mock.return_value = [mock_credential_info]
            download_mock.return_value = None
            get_list_mock.return_value = []
            databricks_artifact_repo.download_artifacts(remote_file_path, local_path)
            read_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=[remote_file_path]
            )
            download_mock.assert_called_with(
                cloud_credential_info=mock_credential_info,
                dst_local_file_path=ANY,
            )

    @pytest.mark.parametrize(
        ("remote_file_path", "local_path"),
        [("test_file.txt", ""), ("test_file.txt", None)],
    )
    def test_databricks_download_file_with_relative_path(
        self, remote_file_path, local_path
    ):
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_run_artifact_root"
        ) as get_run_artifact_root_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_read_credential_infos"
        ) as read_credential_infos_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + ".list_artifacts"
        ) as get_list_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._download_from_cloud"
        ) as download_mock:
            get_run_artifact_root_mock.return_value = MOCK_RUN_ROOT_URI
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_AZURE_SIGNED_URI,
                type=ArtifactCredentialType.AZURE_SAS_URI,
            )
            read_credential_infos_mock.return_value = [mock_credential_info]
            download_mock.return_value = None
            get_list_mock.return_value = []
            databricks_artifact_repo = get_artifact_repository(MOCK_SUBDIR_ROOT_URI)
            databricks_artifact_repo.download_artifacts(remote_file_path, local_path)
            read_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID,
                paths=[posixpath.join(MOCK_SUBDIR, remote_file_path)],
            )
            download_mock.assert_called_with(
                cloud_credential_info=mock_credential_info,
                dst_local_file_path=ANY,
            )

    def test_databricks_download_file_get_request_fail(
        self, databricks_artifact_repo, test_file
    ):
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_read_credential_infos"
        ) as read_credential_infos_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + ".list_artifacts"
        ) as get_list_mock, mock.patch(
            "requests.get"
        ) as request_mock:
            mock_credential_info = ArtifactCredentialInfo(
                signed_uri=MOCK_AZURE_SIGNED_URI,
                type=ArtifactCredentialType.AZURE_SAS_URI,
            )
            read_credential_infos_mock.return_value = [mock_credential_info]
            get_list_mock.return_value = []
            request_mock.return_value = MlflowException("MOCK ERROR")
            with pytest.raises(MlflowException, match=r".+"):
                databricks_artifact_repo.download_artifacts(test_file.strpath)
            read_credential_infos_mock.assert_called_with(
                run_id=MOCK_RUN_ID, paths=[test_file.strpath]
            )

    def test_download_artifacts_awaits_download_completion(
        self, databricks_artifact_repo, tmpdir
    ):
        """
        Verifies that all asynchronous artifact downloads are joined before `download_artifacts()`
        returns a result to the caller
        """
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_read_credential_infos"
        ) as read_credential_infos_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + ".list_artifacts"
        ) as get_list_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._download_from_cloud"
        ) as download_mock:
            ArtifactCredentialInfo(
                signed_uri=MOCK_AZURE_SIGNED_URI,
                type=ArtifactCredentialType.AZURE_SAS_URI,
            )
            read_credential_infos_mock.return_value = [
                ArtifactCredentialInfo(
                    signed_uri=MOCK_AZURE_SIGNED_URI,
                    type=ArtifactCredentialType.AZURE_SAS_URI,
                ),
            ]
            get_list_mock.return_value = [
                FileInfo(path="file_1.txt", is_dir=False, file_size=100),
                FileInfo(path="file_2.txt", is_dir=False, file_size=0),
            ]

            def mock_download_from_cloud(
                cloud_credential_info, dst_local_file_path
            ):  # pylint: disable=unused-argument
                # Sleep in order to simulate a longer-running asynchronous download
                time.sleep(2)
                with open(dst_local_file_path, "w") as f:
                    f.write("content")

            download_mock.side_effect = mock_download_from_cloud

            databricks_artifact_repo.download_artifacts("test_path", str(tmpdir))

            expected_file1_path = os.path.join(str(tmpdir), "file_1.txt")
            expected_file2_path = os.path.join(str(tmpdir), "file_2.txt")
            for path in [expected_file1_path, expected_file2_path]:
                assert os.path.exists(path)
                with open(path, "r") as f:
                    assert f.read() == "content"

    def test_artifact_logging(self, databricks_artifact_repo, tmpdir):
        """
        Verifies that `log_artifact()` and `log_artifacts()` initiate all expected asynchronous
        artifact uploads and await their completion before returning results to the caller
        """
        src_dir = os.path.join(str(tmpdir), "src")
        os.makedirs(src_dir)
        src_file1_path = os.path.join(src_dir, "file_1.txt")
        with open(src_file1_path, "w") as f:
            f.write("file1")
        src_file2_path = os.path.join(src_dir, "file_2.txt")
        with open(src_file2_path, "w") as f:
            f.write("file2")

        dst_dir = os.path.join(str(tmpdir), "dst")
        os.makedirs(dst_dir)

        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credential_infos_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._upload_to_cloud"
        ) as upload_mock:
            write_credential_infos_mock.return_value = [
                ArtifactCredentialInfo(
                    signed_uri=MOCK_AZURE_SIGNED_URI,
                    type=ArtifactCredentialType.AZURE_SAS_URI,
                ),
                ArtifactCredentialInfo(
                    signed_uri=MOCK_AZURE_SIGNED_URI,
                    type=ArtifactCredentialType.AZURE_SAS_URI,
                ),
            ]

            def mock_upload_to_cloud(
                cloud_credential_info, src_file_path, dst_run_relative_artifact_path
            ):  # pylint: disable=unused-argument
                # Sleep in order to simulate a longer-running asynchronous upload
                time.sleep(2)
                dst_run_relative_artifact_path = os.path.join(
                    dst_dir, dst_run_relative_artifact_path
                )
                os.makedirs(
                    os.path.dirname(dst_run_relative_artifact_path), exist_ok=True
                )
                shutil.copyfile(src=src_file_path, dst=dst_run_relative_artifact_path)

            upload_mock.side_effect = mock_upload_to_cloud

            databricks_artifact_repo.log_artifacts(src_dir, "dir_artifact")

            expected_dst_dir_file1_path = os.path.join(
                dst_dir, "dir_artifact", "file_1.txt"
            )
            expected_dst_dir_file2_path = os.path.join(
                dst_dir, "dir_artifact", "file_2.txt"
            )
            assert os.path.exists(expected_dst_dir_file1_path)
            assert os.path.exists(expected_dst_dir_file2_path)
            with open(expected_dst_dir_file1_path, "r") as f:
                assert f.read() == "file1"
            with open(expected_dst_dir_file2_path, "r") as f:
                assert f.read() == "file2"

            databricks_artifact_repo.log_artifact(src_file1_path)

            expected_dst_file_path = os.path.join(dst_dir, "file_1.txt")
            assert os.path.exists(expected_dst_file_path)
            with open(expected_dst_file_path, "r") as f:
                assert f.read() == "file1"

    def test_download_artifacts_provides_failure_info(self, databricks_artifact_repo):
        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_read_credential_infos"
        ) as read_credentials_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + ".list_artifacts"
        ) as get_list_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._download_from_cloud"
        ) as download_mock:
            get_list_mock.return_value = [
                FileInfo(path="file_1.txt", is_dir=False, file_size=100),
                FileInfo(path="file_2.txt", is_dir=False, file_size=0),
            ]
            read_credentials_mock.return_value = [
                ArtifactCredentialInfo(
                    signed_uri=MOCK_AZURE_SIGNED_URI,
                    type=ArtifactCredentialType.AZURE_SAS_URI,
                ),
            ]
            download_mock.side_effect = [
                MlflowException("MOCK ERROR 1"),
                MlflowException("MOCK ERROR 2"),
            ]

            match = (
                r"The following failures occurred while downloading one or more artifacts.+"
                r"MOCK ERROR 1.+"
                r"MOCK ERROR 2"
            )
            with pytest.raises(MlflowException, match=match) as exc:
                databricks_artifact_repo.download_artifacts("test_path")

            err_msg = str(exc.value)
            assert MOCK_RUN_ROOT_URI in err_msg
            assert "file_1.txt" in err_msg
            assert "MOCK ERROR 1" in err_msg
            assert "file_2.txt" in err_msg
            assert "MOCK ERROR 2" in err_msg

    def test_log_artifacts_provides_failure_info(
        self, databricks_artifact_repo, tmpdir
    ):
        src_file1_path = os.path.join(str(tmpdir), "file_1.txt")
        with open(src_file1_path, "w") as f:
            f.write("file1")
        src_file2_path = os.path.join(str(tmpdir), "file_2.txt")
        with open(src_file2_path, "w") as f:
            f.write("file2")

        with mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._get_write_credential_infos"
        ) as write_credentials_mock, mock.patch(
            DATABRICKS_ARTIFACT_REPOSITORY + "._upload_to_cloud"
        ) as upload_mock:
            write_credentials_mock.return_value = [
                ArtifactCredentialInfo(
                    signed_uri=MOCK_AZURE_SIGNED_URI,
                    type=ArtifactCredentialType.AZURE_SAS_URI,
                ),
                ArtifactCredentialInfo(
                    signed_uri=MOCK_AZURE_SIGNED_URI,
                    type=ArtifactCredentialType.AZURE_SAS_URI,
                ),
            ]
            upload_mock.side_effect = [
                MlflowException("MOCK ERROR 1"),
                MlflowException("MOCK ERROR 2"),
            ]
            match = (
                r"The following failures occurred while uploading one or more artifacts.+"
                r"MOCK ERROR 1.+"
                r"MOCK ERROR 2"
            )
            with pytest.raises(MlflowException, match=match) as exc:
                databricks_artifact_repo.log_artifacts(str(tmpdir), "test_artifacts")

            err_msg = str(exc.value)
            assert MOCK_RUN_ROOT_URI in err_msg
            assert "file_1.txt" in err_msg
            assert "MOCK ERROR 1" in err_msg
            assert "file_2.txt" in err_msg
            assert "MOCK ERROR 2" in err_msg



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="mlflow-databricks-artifacts",
    version="2.0.1",
    description="Plugin to create and access MLflow-managed artifacts on Databricks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Databricks",
    packages=find_packages(exclude=["tests", "tests.*"]),
    install_requires=[
        "mlflow-skinny>1.10.0",
    ],
    entry_points={
        "mlflow.artifact_repository": [
            "dbfs=mlflow_databricks_artifacts.store.entrypoint:dbfs_artifact_repo_factory",
        ]
    },
)



import json
import os
import posixpath
from distutils.version import LooseVersion

import mlflow
from mlflow.exceptions import MlflowException

try:
    from mlflow.store.artifact.dbfs_artifact_repo import DbfsRestArtifactRepository
    from mlflow.store.artifact.local_artifact_repo import LocalArtifactRepository
except ImportError:
    # Older versions of MLflow placed artifact repository implementations
    # in the top-level `mlflow.store` module, rather than the
    # `mlflow.store.artifact` module; `DbfsArtifactRepository` was also
    # renamed to `DbfsRestArtifactRepository`
    from mlflow.store.dbfs_artifact_repo import (
        DbfsArtifactRepository as DbfsRestArtifactRepository,
    )
    from mlflow.store.local_artifact_repo import LocalArtifactRepository


USE_FUSE_ENV_VAR = "MLFLOW_ENABLE_DBFS_FUSE_ARTIFACT_REPO"


def dbfs_artifact_repo_factory(artifact_uri):
    """
    Returns an ArtifactRepository subclass for storing artifacts on DBFS.

    This factory method is used with URIs of the form ``dbfs:/<path>``. DBFS-backed artifact
    storage can only be used together with the RestStore.

    In the special case where the URI is of the form
    `dbfs:/databricks/mlflow-tracking/<Exp-ID>/<Run-ID>/<path>',
    a DatabricksArtifactRepository is returned. This is capable of storing access controlled
    artifacts.

    :param artifact_uri: DBFS root artifact URI (string).
    :return: Subclass of ArtifactRepository capable of storing artifacts on DBFS.
    """
    try:
        if not supported_mlflow_version(mlflow.__version__):
            from mlflow.store.artifact.dbfs_artifact_repo import (
                dbfs_artifact_repo_factory,
            )

            return dbfs_artifact_repo_factory(artifact_uri)
    except Exception:
        pass

    # For some reason, we must import modules specific to this package within the
    # entrypoint function rather than the top-level module. Otherwise, entrypoint
    # registration fails with import errors
    from mlflow_databricks_artifacts.store.artifact_repo import (
        DatabricksArtifactRepository,
    )
    from mlflow_databricks_artifacts.utils.databricks_utils import (
        is_dbfs_fuse_available,
    )
    from mlflow_databricks_artifacts.utils.string_utils import strip_prefix
    from mlflow_databricks_artifacts.utils.uri import (
        get_databricks_profile_uri_from_artifact_uri,
        is_databricks_acled_artifacts_uri,
        is_databricks_model_registry_artifacts_uri,
        is_valid_dbfs_uri,
        remove_databricks_profile_info_from_artifact_uri,
    )

    if not is_valid_dbfs_uri(artifact_uri):
        raise MlflowException(
            "DBFS URI must be of the form dbfs:/<path> or "
            + "dbfs://profile@databricks/<path>, but received "
            + artifact_uri
        )

    cleaned_artifact_uri = artifact_uri.rstrip("/")
    db_profile_uri = get_databricks_profile_uri_from_artifact_uri(cleaned_artifact_uri)
    if is_databricks_acled_artifacts_uri(artifact_uri):
        return DatabricksArtifactRepository(cleaned_artifact_uri)
    elif (
        is_dbfs_fuse_available()
        and os.environ.get(USE_FUSE_ENV_VAR, "").lower() != "false"
        and not is_databricks_model_registry_artifacts_uri(artifact_uri)
        and (db_profile_uri is None or db_profile_uri == "databricks")
    ):
        # If the DBFS FUSE mount is available, write artifacts directly to
        # /dbfs/... using local filesystem APIs.
        # Note: it is possible for a named Databricks profile to point to the current workspace,
        # but we're going to avoid doing a complex check and assume users will use `databricks`
        # to mean the current workspace. Using `DbfsRestArtifactRepository` to access the current
        # workspace's DBFS should still work; it just may be slower.
        final_artifact_uri = remove_databricks_profile_info_from_artifact_uri(
            cleaned_artifact_uri
        )
        file_uri = "file:///dbfs/{}".format(strip_prefix(final_artifact_uri, "dbfs:/"))
        return LocalArtifactRepository(file_uri)
    return DbfsRestArtifactRepository(cleaned_artifact_uri)


def supported_mlflow_version(mlflow_version):
    """
    :return: `True` if the given version of MLflow is supported by the plugin
    """
    mlflow_version = LooseVersion(mlflow_version)
    return LooseVersion("1.10.0") < mlflow_version <= LooseVersion("1.29.0")
